#pragma once
#include <map>
#include <memory>

template <class FastData, class ColdData>
class OutOfLine {
  inline static std::map<OutOfLine const*, std::unique_ptr<ColdData>> global_map_;
 public:
  template <class... TArgs>
  OutOfLine(TArgs&&... args) : OutOfLine(TwoPhaseInit()) {
    init_cold_data(std::forward<TArgs>(args)...);
  }
  ~OutOfLine() { global_map_.erase(this); }
  explicit OutOfLine(OutOfLine&& other) : OutOfLine(OutOfLine::TwoPhaseInit()) { (*this) = std::move(other); }
  OutOfLine& operator=(OutOfLine&& other) {
    global_map_[this] = std::move(global_map_[&other]);
    return *this;
  }
  OutOfLine(OutOfLine const&) = delete;
  OutOfLine& operator=(OutOfLine const&) = delete;

  ColdData& cold() noexcept { return *global_map_[this]; }
  ColdData const& cold() const noexcept { return *global_map_[this]; }

  struct TwoPhaseInit {};
  OutOfLine(TwoPhaseInit) { global_map_.try_emplace(this); }
  template <class... TArgs>
  void init_cold_data(TArgs&&... args) {
    global_map_.find(this)->second = std::make_unique<ColdData>(std::forward<TArgs>(args)...);
  }

  void release_cold_data() { global_map_[this].reset(); }
};
